
# Measurement intervals, peak intensity values, and uncertainty for each
# point
h2oex.times <- %s/1000.0
h2oex.vals  <- %s
h2oex.err   <- %.5f


# Water relaxation and delta-F
h2o.R  <- %.5f
h2o.dF <- %.5f

# Assignment ID
ass <- "%s"
res <- %i
alt <- "%s"

# Output
plot.file <- "%s"
fit.file  <- "%s"

# Initial guess for optimization (1H R1, kex)
initial.par <- c(0.8, 2.5)

# Limits for plotting
t1.plot.xlim <- c(0, 5)
t1.plot.ylim <- c(-1.1, 1.1)

h2oex.plot.xlim <- c(0, 5)
h2oex.plot.ylim <- c(0.0, 1.1*h2o.dF)

# Number of simulations (bootstrap or otherwise) to perform
num.sim <- 1000

# Method to use when estimating errors - can be one of:
#
# hessian  - Calculate errors based on inverse of hessian matrix.
#            Fast, but the matrix could be singular.
# hess-fit - Calculate errors based on hessian matrix, but use
#            residuals to estimate the uncertainty instead of h2oex.err
# mc       - Do a monte-carlo simulation based on original points
#            +/- h2oex.err
# mc-sd    - Do a monte-carlo simulation based on original points
#            +/- residual to best fit
# mc-fit   - Do a monte-carlo simulation based on best fit points
#            +/- residual to original points
# mc-scale - Do a monte-carlo simulation based on best fit points
#            +/- residual to original points, scale residuals until
#            avg SSR is comparable to the original SSR.
# boot     - Resample the original points with replacement and repeat
#            the fit
# jack     - Resample the original points, leaving one out each time,
#          - and repeat the fit

err.method <- 'mc-scale'


##############################################################################
## Functions
##############################################################################

library(boot)

# Useful function for plotting error bars
err.bars <- function (x, y, ebl, ebu = ebl, length = 0.00, ...)
  arrows(x, y + ebu, x, y - ebl, angle = 90, code = 3,
         length = length, ...)

# Fitting T1 data ############################################################

# Plot T1 (apparent) given R, kex
t1.plot <- function(par) {
  R <- par[1]
  k <- par[2]
  x <- seq(t1.plot.xlim[1], t1.plot.xlim[2], length.out=100)
  y <- 2.0*exp( -(R+k)*x ) - 1.0

  data.frame(x,y)
}

# Calculate sum of squared residuals given R, kex
t1.ssr <- function(par) {
  R <- par[1]
  k <- par[2]

  # Ignore first point; it was used to establish +/- 1.0 bounds.

  x  <- t1.times[2:length(t1.times)]
  y  <- t1.vals[2:length(t1.vals)]
  fx <- 2.0*exp( -(R+k)*x ) - 1.0

  ssr <- sum((y-fx)*(y-fx))
  ssr
}

# Fitting h2o exchange data #################################################

# Plot the intensity given R, kex
h2oex.plot <- function(par) {
  R <- par[1]
  k <- par[2]
  x <- seq(h2oex.plot.xlim[1], h2oex.plot.xlim[2], length.out=100)
# y <- 1.0 - 2.0*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))
  y <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

  data.frame(x,y)
}

h2oex.func <- function(par) {
  R <- par[1]
  k <- par[2]
# y <- 1.0 - 2.0*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

  f <- function(x) {
    h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))
  }

  f
}

# Calculate sum of squared residuals given R, kex
h2oex.ssr <- function(par) {
  R <- par[1]
  k <- par[2]

  # Ignore first point; it was used to establish +/- 1.0 bounds.

  x  <- h2oex.times[2:length(h2oex.times)]
  y  <- h2oex.vals[2:length(h2oex.vals)]
  fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

  ssr <- sum((y-fx)*(y-fx))
  ssr
}

h2oex.J <- function(par) {
  R <- par[1]
  k <- par[2]

  # Ignore first point; it was used to establish +/- 1.0 bounds.

  x  <- h2oex.times[2:length(h2oex.times)]
  y  <- h2oex.vals[2:length(h2oex.vals)]
  fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

  ssr <- 0.5*sum((y-fx)*(y-fx)/h2oex.err/h2oex.err)
  ssr
}

# Calculate correlation coefficient (pearson r) given R, kex
h2oex.cor <- function(par) {
  R <- par[1]
  k <- par[2]

  # Ignore first point; it was used to establish +/- 1.0 bounds.

  x  <- h2oex.times[2:length(h2oex.times)]
  y  <- h2oex.vals[2:length(h2oex.vals)]
# fx <- 1.0 - 2.0*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))
  fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))
  cor(y, fx)
}

# Plot water R1 relaxation
h2or1.plot <- function() {
  x <- seq(h2oex.plot.xlim[1], h2oex.plot.xlim[2], length.out=100)
  y <- h2o.dF*exp( - h2o.R*x )

  data.frame(x,y)
}

# Bootstrap fitting function - given a resampled data set and initial
# paramters, determine the best fit parameters
h2oex.fit.func <- function(data, par.best) {
  x <- data$x
  y <- data$y
  
  ssr <- function(par) {
    R <- par[1]
    k <- par[2]
    
#   fx <- 1.0 - 2.0*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))
    fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))
    sum( (fx-y)*(fx-y) )
  }

  opt <- optim(par.best, ssr, method="L-BFGS-B", lower=c(0, 0))

  c(opt$par)
}

# Function to apply resampled data to fitting function above, idx is
# resampled indices for data
stat.func <- function(data, idx, par.best) h2oex.fit.func(data[idx,], par.best)

##############################################################################
## Fitting section
##############################################################################

# Find best parameters

opt.list <- optim(initial.par, h2oex.ssr, method="L-BFGS-B",
                  lower=c(0, 0))
par.best <- c(opt.list$par)
r        <- h2oex.cor(opt.list$par)

if (err.method == 'boot') {
  # Set up bootstrap data structure
  tmp.x  <- h2oex.times[2:length(h2oex.times)]     # Ignore first point, as 
  tmp.y  <- h2oex.vals[2:length(h2oex.vals)]       # before
  boot.xy <- data.frame(cbind(x=tmp.x, y=tmp.y))

  # Bootstrap to get errors
  boot.fit <- boot(boot.xy, stat.func, num.sim, par.best=par.best)
  errs <- c(sd(boot.fit$t[,1]), sd(boot.fit$t[,2]))

} else if (err.method == 'hessian') {
  opt.list.chi <- optim(initial.par, h2oex.J, method="L-BFGS-B",
                        hessian=T, lower=c(0, 0))
  errs <- sqrt(diag(solve(opt.list.chi$hessian)))

} else if (err.method == 'hess-fit') {
  h2oex.calc <- apply(array(h2oex.times), 1, h2oex.func(par.best))
  h2oex.calc[1] <- 1.0
  diff <- abs(h2oex.calc - h2oex.vals)

  h2oex.J.fit <- function(par) {
    R <- par[1]
    k <- par[2]
    
    # Ignore first point; it was used to establish +/- 1.0 bounds.
    
    x  <- h2oex.times[2:length(h2oex.times)]
    y  <- h2oex.vals[2:length(h2oex.vals)]
    fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

    ssr <- 0.5*sum((y-fx)*(y-fx)/diff/dfff)
    ssr
  }
  
  opt.list.chi <- optim(initial.par, h2oex.J, method="L-BFGS-B",
                        hessian=T, lower=c(0, 0))
  errs <- sqrt(diag(solve(opt.list.chi$hessian)))

} else if (err.method == 'jack') {
  k.obs.sim <- array(dim=length(h2oex.times)-1)
  r1.sim    <- array(dim=length(h2oex.times)-1)

  for (i in seq(length(h2oex.times)-1)) {
    initial.mc <- par.best

    if (i == length(h2oex.times)-1) {
      h2oex.vals.mc <- h2oex.vals[c(1:i)]
      h2oex.time.mc <- h2oex.times[c(1:i)]
    } else {
      h2oex.vals.mc <- h2oex.vals[c(1:i, seq(2+i, length(h2oex.vals)))]
      h2oex.time.mc <- h2oex.times[c(1:i, seq(2+i, length(h2oex.vals)))]
    }

    h2oex.vals.mc[1] <- 1.0

    #print(h2oex.vals)
    #print(h2oex.time.mc)
    #print(h2oex.vals.mc)
    
    h2oex.ssr.mc <- function(par) {
      R <- par[1]
      k <- par[2]

      # Ignore first point; it was used to establish +/- 1.0 bounds.
      x  <- h2oex.time.mc[2:length(h2oex.time.mc)]
      y  <- h2oex.vals.mc[2:length(h2oex.vals.mc)]
      fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

      ssr <- sum((y-fx)*(y-fx))
      ssr
    }

    opt.list.mc <- optim(initial.mc, h2oex.ssr.mc, method="L-BFGS-B",
                         lower=c(0,0))

    par.best.mc  <- c(opt.list.mc$par)
    #print(h2oex.mc)
    #print(c(i, par.best.mc[1], par.best.mc[2]))

    r1.sim[i]    <- par.best.mc[1]
    k.obs.sim[i] <- par.best.mc[2]
  }

  errs <- c(sd(r1.sim), sd(k.obs.sim))
  
} else if (err.method == 'mc') {
  k.obs.sim <- array(dim=num.sim)
  r1.sim    <- array(dim=num.sim)

  for (i in seq(num.sim)) {
    initial.mc <- par.best
    h2oex.mc <- rnorm(length(h2oex.vals), mean=h2oex.vals, sd=h2oex.err)
    h2oex.mc[1] <- 1.0
    
    h2oex.ssr.mc <- function(par) {
      R <- par[1]
      k <- par[2]

      # Ignore first point; it was used to establish +/- 1.0 bounds.
      x  <- h2oex.times[2:length(h2oex.times)]
      y  <- h2oex.mc[2:length(h2oex.mc)]
      fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

      ssr <- sum((y-fx)*(y-fx))
      ssr
    }

    opt.list.mc <- optim(initial.mc, h2oex.ssr.mc, method="L-BFGS-B",
                         lower=c(0,0))

    par.best.mc  <- c(opt.list.mc$par)
    #print(h2oex.mc)
    #print(c(i, par.best.mc[1], par.best.mc[2]))

    r1.sim[i]    <- par.best.mc[1]
    k.obs.sim[i] <- par.best.mc[2]
  }

  errs <- c(sd(r1.sim), sd(k.obs.sim))

} else if (err.method == 'mc-sd') {
  k.obs.sim <- array(dim=num.sim)
  r1.sim    <- array(dim=num.sim)

  h2oex.calc <- apply(array(h2oex.times), 1, h2oex.func(par.best))
  h2oex.calc[1] <- 1.0
  diff <- abs(h2oex.calc - h2oex.vals)

  #print(h2oex.vals)
  #print(h2oex.calc)
  #print(diff)
  
  for (i in seq(num.sim)) {
    initial.mc <- par.best
    
    h2oex.mc <- rnorm(length(h2oex.vals), mean=h2oex.vals, sd=diff)
    h2oex.mc[1] <- 1.0
    
    h2oex.ssr.mc <- function(par) {
      R <- par[1]
      k <- par[2]

      # Ignore first point; it was used to establish +/- 1.0 bounds.
      x  <- h2oex.times[2:length(h2oex.times)]
      y  <- h2oex.mc[2:length(h2oex.mc)]
      fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

      ssr <- sum((y-fx)*(y-fx))
      ssr
    }

    opt.list.mc <- optim(initial.mc, h2oex.ssr.mc, method="L-BFGS-B",
                         lower=c(0,0))

    par.best.mc  <- c(opt.list.mc$par)
    #print(h2oex.mc)
    #print(c(i, par.best.mc[1], par.best.mc[2]))

    r1.sim[i]    <- par.best.mc[1]
    k.obs.sim[i] <- par.best.mc[2]
  }

  errs <- c(sd(r1.sim), sd(k.obs.sim))

} else if (err.method == 'mc-fit') {
  k.obs.sim <- array(dim=num.sim)
  r1.sim    <- array(dim=num.sim)

  h2oex.calc <- apply(array(h2oex.times), 1, h2oex.func(par.best))
  h2oex.calc[1] <- 1.0
  diff <- abs(h2oex.calc - h2oex.vals)

  #print(h2oex.vals)
  #print(h2oex.calc)
  #print(diff)
  
  for (i in seq(num.sim)) {
    initial.mc <- par.best
    
    h2oex.mc <- rnorm(length(h2oex.vals), mean=h2oex.calc, sd=diff)
    h2oex.mc[1] <- 1.0
    
    h2oex.ssr.mc <- function(par) {
      R <- par[1]
      k <- par[2]

      # Ignore first point; it was used to establish +/- 1.0 bounds.
      x  <- h2oex.times[2:length(h2oex.times)]
      y  <- h2oex.mc[2:length(h2oex.mc)]
      fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

      ssr <- sum((y-fx)*(y-fx))
      ssr
    }

    opt.list.mc <- optim(initial.mc, h2oex.ssr.mc, method="L-BFGS-B",
                         lower=c(0,0))

    par.best.mc  <- c(opt.list.mc$par)
    #print(h2oex.mc)
    #print(c(i, par.best.mc[1], par.best.mc[2]))

    r1.sim[i]    <- par.best.mc[1]
    k.obs.sim[i] <- par.best.mc[2]
  }

  errs <- c(sd(r1.sim), sd(k.obs.sim))

} else if (err.method == 'mc-scale') {
  k.obs.sim <- array(dim=num.sim)
  r1.sim    <- array(dim=num.sim)
  ssr.sim   <- array(dim=num.sim)
  
  h2oex.calc <- apply(array(h2oex.times), 1, h2oex.func(par.best))
  h2oex.calc[1] <- 1.0
  diff <- abs(h2oex.calc - h2oex.vals)

  #print(h2oex.vals)
  #print(h2oex.calc)
  #print(diff)
  
  # First pass, calculate the "unscaled" ssr
  for (i in seq(num.sim)) {
    initial.mc <- par.best
    
    h2oex.mc <- rnorm(length(h2oex.vals), mean=h2oex.calc, sd=diff)
    h2oex.mc[1] <- 1.0
    
    h2oex.ssr.mc <- function(par) {
      R <- par[1]
      k <- par[2]

      # Ignore first point; it was used to establish +/- 1.0 bounds.
      x  <- h2oex.times[2:length(h2oex.times)]
      y  <- h2oex.mc[2:length(h2oex.mc)]
      fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

      ssr <- sum((y-fx)*(y-fx))
      ssr
    }

    opt.list.mc <- optim(initial.mc, h2oex.ssr.mc, method="L-BFGS-B",
                         lower=c(0,0))

    par.best.mc  <- c(opt.list.mc$par)
    #print(h2oex.mc)
    #print(c(i, par.best.mc[1], par.best.mc[2]))

    r1.sim[i]    <- par.best.mc[1]
    k.obs.sim[i] <- par.best.mc[2]
    ssr.sim[i]   <- h2oex.ssr.mc(opt.list.mc$par)
  }

  print(c(mean(ssr.sim), sd(ssr.sim), h2oex.ssr(par.best)))
  scale <- h2oex.ssr(par.best)/mean(ssr.sim)
  
  # Second pass: now scale the SSR
  k.obs.sim <- array(dim=num.sim)
  r1.sim    <- array(dim=num.sim)
  ssr.sim   <- array(dim=num.sim)
  
  for (i in seq(num.sim)) {
    initial.mc <- par.best
    
    h2oex.mc <- rnorm(length(h2oex.vals), mean=h2oex.calc, sd=diff*scale)
    h2oex.mc[1] <- 1.0
    
    h2oex.ssr.mc <- function(par) {
      R <- par[1]
      k <- par[2]

      # Ignore first point; it was used to establish +/- 1.0 bounds.
      x  <- h2oex.times[2:length(h2oex.times)]
      y  <- h2oex.mc[2:length(h2oex.mc)]
      fx <- h2o.dF*k/(R+k-h2o.R) * (exp(-(h2o.R*x)) - exp(-(R+k)*x))

      ssr <- sum((y-fx)*(y-fx))
      ssr
    }

    opt.list.mc <- optim(initial.mc, h2oex.ssr.mc, method="L-BFGS-B",
                         lower=c(0,0))

    par.best.mc  <- c(opt.list.mc$par)
    #print(h2oex.mc)
    #print(c(i, par.best.mc[1], par.best.mc[2]))

    r1.sim[i]    <- par.best.mc[1]
    k.obs.sim[i] <- par.best.mc[2]
    ssr.sim[i]   <- h2oex.ssr.mc(opt.list.mc$par)
  }

  print(c(mean(ssr.sim), sd(ssr.sim), h2oex.ssr(par.best)))
  errs <- c(sd(r1.sim), sd(k.obs.sim))
}

##############################################################################
## Plotting/Output
##############################################################################

#plot(NA, NA, xlim=t1.plot.xlim, ylim=t1.plot.ylim, xlab='Time (s)',
#     ylab='Relative Intensity',
#     main=paste("1H T1 Relaxation (", ass, ")", sep=''))

#points(t1.times, t1.vals)
#err.bars(t1.times, t1.vals, t1.err)

postscript(plot.file, height=7.5, width=7.5)

plot(NA, NA, xlim=h2oex.plot.xlim, ylim=h2oex.plot.ylim, xlab='Time (s)',
     ylab='Relative Intensity',
     main=sprintf("H20 Exchange (%%s)\nR^2 = %%.4f R1 = %%.3f (%%.3f) kex = %%.3f (%%.3f)",
       ass, r*r, par.best[1], errs[1], par.best[2], errs[2]))

nt <- length(h2oex.times)
points(h2oex.times[2:nt], h2oex.vals[2:nt])
err.bars(h2oex.times[2:nt], h2oex.vals[2:nt], h2oex.err)
lines(h2or1.plot(), col='blue', lty=2)
lines(h2oex.plot(opt.list$par), col='red')

dev.off()

out <- sprintf("%%5i %%2s %%10.5f %%10.5f %%10.5f %%10.5f %%10.5f",
               res, alt, par.best[1], errs[1],
               par.best[2], errs[2], r*r)

cat(out, file=fit.file)
