#!/usr/local/bin/python

#
# title: fit_all.py
# summary: H2O Exchange fitting script
# author:  Nicholas Fitzkee (nfitzkee at nih dot gov)
# date:    August 16, 2010
#

#
# Parameters - Change these values as needed
###########################################################################

#
# Parameters from the pulse program setup: the table of intensities,
# the time points, and the noise level for estimating uncertainties
#

h2oex_table = '../h2oex/axt.tab'
h2oex_noise = 13279
h2oex_times = [0, 20, 60, 100, 180, 240, 360, 480]

#
# Parameters for water: the delta_f value (f[+] - f[-]) and the water
# R1 value
#

delta_f = 1.90
h2o_R1  = 1.0/3.00

#
# Correction values: use to normalize H2O exchange values relative
# to the reference experiment, e.g. if sample degradation occurs.
# Set correction_tab to None if not used.
#

#correction_tab = '../correction/axt.tab'
correction_tab = None
default_correction = 1.344

#
# Output directories for plotting, for fit data, for scripts, and for
# R program ouput.
#

plot_dir  = './plot'
fit_dir   = './fit'
scr_dir   = './script'
out_dir   = './out'
curve_dir = './gfx_data'

#
# File in which to write summary data.
#

summary  = './summary.txt'

#
###########################################################################

R_SCRIPT = "fit_T1kex.R"

import gdb, biotools, os, shutil, math
h2gdb = gdb.GDBTab(h2oex_table)

#
# Helper functions
###########################################################################

def index(tab):
    result = {}
    keys = {}

    for ent in tab:
        try:
            r1, r2 = biotools.parse_sparky2(ent['ASS'])
        except:
            print 'Unable to parse assignment: %s' % ent['ASS']
            continue

        if keys.has_key((r2.index, r2.alt)):
            print 'Ignoring duplicate entry: %s' % ent['ASS']
            continue

        result[(r2.index, r2.alt)] = ent
        keys[(r2.index, r2.alt)] = 1

    keys = keys.keys()
    keys.sort()
    return keys, result

#
# Begin main program script
###########################################################################

h2keys, h2idx = index(h2gdb)

corkeys, coridx = [], {}
if correction_tab and os.path.exists(correction_tab):
    corgdb = gdb.GDBTab(correction_tab)
    corkeys, coridx = index(corgdb)

rfile = open(R_SCRIPT)
script = rfile.read()
rfile.close()

sfile = open(summary, 'w')

# Remove any stale fitting files
for d in (plot_dir, scr_dir, fit_dir, out_dir, curve_dir):
    if os.path.exists(d):
        shutil.rmtree(d)
    os.makedirs(d)


# Main fitting loop:
for res in h2keys:
    ridx, alt = res

#    if ridx > 53: break
    
    altstr = alt
    if alt == None:
        altstr = '_'
    
    print '* Fitting residue %i%s (%s)...' % (ridx, altstr, h2idx[res]['ASS'])

    # Apply correction factor if needed
    fac = 1.0
    if correction_tab:
        #fac = coridx.get(res, default_correction)
        if coridx.has_key(res):
            print '  - Correction factor: %.3f' % coridx[res]['Z_A1']
            fac = coridx[res]['Z_A1']
        else:
            print '  - Default correction factor: %.3f' % default_correction
            fac = default_correction
        
    h2vals = []
    for i in xrange(len(h2oex_times)):
        if i == 0:
            h2vals.append(h2idx[res]['Z_A%i' % i])
        else:
            h2vals.append(h2idx[res]['Z_A%i' % i]/fac)
        
    h2err = abs(h2oex_noise/h2idx[res]['HEIGHT'])

    h2times_str = 'c(%s)' % ', '.join(['%.4f' % x for x in h2oex_times])
    h2vals_str  = 'c(%s)' % ', '.join(['%.4f' % x for x in h2vals])

    ass = h2idx[res]['ASS']

    my_script = '%s/script-%i%s.R' % (scr_dir, ridx, altstr)
    plotfile  = '%s/plot-%i%s.ps' % (plot_dir, ridx, altstr) 
    fitfile   = '%s/fit-%i%s.txt' % (fit_dir, ridx, altstr)
    outfile   = '%s/out-%i%s.txt' % (out_dir, ridx, altstr)

    pntsfile  = '%s/gfx-%i%s.pts.dat' % (curve_dir, ridx, altstr)
    curvefile = '%s/gfx-%i%s.fit.dat' % (curve_dir, ridx, altstr)
    h2ofile   = '%s/gfx-%i%s.h2o.dat' % (curve_dir, ridx, altstr)

    # Write the data into the R script for fitting
    f = open(my_script, 'w')
    f.write(script % (h2times_str,
                      h2vals_str,
                      h2err,
                      h2o_R1,
                      delta_f,
                      ass,
                      ridx,
                      altstr,
                      plotfile,
                      fitfile
                      )
                      
            )
    f.close()

    # Run R
    os.system("R --vanilla < %s > %s 2>&1" % (my_script, outfile))

    f = open(fitfile)
    l = f.readline().strip()
    f.close()

    print l
    sfile.write('%s\n' % l)
    sfile.flush()

    f = open(pntsfile, 'w')
    for i in xrange(1, len(h2oex_times)):
        f.write('%f %f %f\n' % (h2oex_times[i]/1000.0, h2vals[i], h2err))
    f.close()

    toks = l.split()
    knoe = float(toks[4])
    r1   = float(toks[2])
    
    itime = 0.0
    ftime = 4.0
    dtime = 0.001

    f1 = open(curvefile, 'w')
    f2 = open(h2ofile, 'w')
    
    while itime < ftime:
        y = delta_f*knoe/(r1+knoe-h2o_R1) * \
          (math.exp(-(h2o_R1*itime)) - math.exp(-(r1+knoe)*itime))
        h2o = delta_f*math.exp(-(h2o_R1*itime))
        
        f1.write('%.6e %.6e\n' % (itime, y))
        f2.write('%.6e %.6e\n' % (itime, h2o))
        
        itime = itime + dtime

    f1.close()
    f2.close()
    
sfile.close()
        
